/*
/* do_job.c copyright (c) 1996 Rob Ross, all rights reserved.
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as published
	by the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	Contact:  Rob Ross  rbross@parl.eng.clemson.edu  or
				 Matt Cettei mcettei@parl.eng.clemson.edu
/*/

/*
/* DO_JOB.C - performs IOD interaction for application tasks
/*
/*/

/*
 * $Log: do_job.c,v $
 * Revision 1.10  1999/06/17 15:36:46  rbross
 * *** empty log message ***
 *
 * Revision 1.9  1999/02/19 19:11:36  rbross
 * better errno passing, cleanup
 *
 * Revision 1.8  1998/07/23 21:56:45  rbross
 * cleanup
 *
 * Revision 1.7  1997/10/27 21:01:57  pvfs
 * added license
 *
 * Revision 1.6  1997/03/19 06:01:08  rbross
 * bug fixes in do_job
 *
 * Revision 1.5  1997/03/18 00:04:31  rbross
 * modified to use a_get_start() to check the head of a job to see if the
 * socket matches instead of looking for the first access for the job
 * with a_search()
 *
 * Revision 1.4  1997/03/13 21:23:21  rbross
 * switched IOD_RW_READ to J_READ
 *
 * Revision 1.3  1997/02/26 15:47:51  rbross
 * *** empty log message ***
 *
 * Revision 1.2  1997/01/14 21:53:17  rbross
 * supports truncating (read) jobs
 *
 * Revision 1.1  1997/01/13 15:53:32  mcettei
 * Initial revision
 *
/*/

/* INCLUDES */
#include <alist.h>
#include <jlist.h>
#include <req.h>
#include <sockio.h>
#include <sockset.h>
#include <debug.h>
#include <prune.h>
#include <errno.h>

extern int errno;

static char do_job_c[] = "$Header: /projects/pvfs/pvfs-working/lib/RCS/do_job.c,v 1.10 1999/06/17 15:36:46 rbross Exp $";

/* DO_JOBS() - 
 * PARAMETERS:
 * jl_p - list of jobs corresponding to sockets in ss_p
 * ss_p - sockset checked for readiness for access
 * msec - time to wait (in msec) before timing out; -1 blocks
 *        indefinitely
 *
 * Returns -1 on error, 0 on success.
 */
int do_jobs(jlist_p jl_p, sockset_p ss_p, int msec)
{
	int s, ret;
	jinfo_p j_p;

	/* NEED TO BE LOOKING FOR ERRORS... */
	if (check_socks(ss_p, msec) < 0) {
		ERR1("do_jobs: check socks failed: %s\n", strerror(errno));
		return(-1); /* wait some amount of time */
	}
	while ((s = nextsock(ss_p)) >= 0) /* still sockets ready */ {
		if (!(j_p = j_search(jl_p, s))) /* no job for socket */ {
			ERR1("do_jobs: no job for %d (so why are we watching it?)\n", s);
			delsock(s, ss_p);
		}
		else {
			if (do_job(s, j_p, ss_p) < 0) {
				ERR1("do_jobs: do_job failed: %s\n", strerror(errno));
				return(-1);
			}
			if (alist_empty(j_p->al_p)) {
				/* remove job from joblist */
				LOG("do_jobs: no more accesses - removing job\n");
				if ((ret = (j_rem(jl_p, s))) < 0) {
					ERR("do_jobs: j_rem failed\n");
				}	
			}	
		}
	}
	return(0);
}

/* returns -1 on error, 0 on partial or full success */
int do_job(int sock, jinfo_p j_p, sockset_p ss_p)
{
	ainfo_p a_p;
	int comp, oldtype, myerr;

	/* find first access in job for socket (if at top of list) */
	if (!(a_p = a_get_start(j_p->al_p))) {
		ERR("do_job: something bad happened finding top access\n");
		delsock(sock, ss_p);
		close(sock);
		errno = EINVAL;
		return(-1);
	}
	if (a_p->sock != sock) /* top access not for this socket */ {
		LOG2("top sock(%d) != current sock(%d)\n", a_p->sock, sock);
		return(0);
	}
	oldtype = a_p->type;

	/* while there are accesses for this socket and we don't block */
	while (a_p) {
		/* try to perform the access (read/write/ack/whatever) */
		switch (a_p->type) {
			case A_READ:
			case A_WRITE:
				LOG3("read/write (s = %d, ad = %lu, sz = %d) ", a_p->sock,
					a_p->u.rw.loc, a_p->u.rw.size);
				if (a_p->type == A_READ) comp = nbrecv(a_p->sock,
					a_p->u.rw.loc, a_p->u.rw.size);
				
				else comp = nbsend(a_p->sock, a_p->u.rw.loc, a_p->u.rw.size);

				if (comp < 0) /* error */ {
					myerr = errno;
					PERROR("do_job: nbsend/nbrecv");
					/* check the error and try to decide what to do */
					a_ptr_rem(j_p->al_p, a_p);
					errno = myerr;
					return(-1);
				}
				else if ((a_p->u.rw.size -= comp) <= 0) /* done with access */ {
					LOG("done!\n");
					a_ptr_rem(j_p->al_p, a_p);
				}
				else /* partially completed */ {
					LOG("partial\n");
					a_p->u.rw.loc  += comp; /* subtract from size above */
					return(0);
				}
				break;
			case A_ACK:
				LOG1("ack (s = %d) ", a_p->sock);
				comp = nbrecv(a_p->sock, a_p->u.ack.cur_p, a_p->u.ack.size);
				if (comp < 0) {
					myerr = errno;
					PERROR("do_job: nbrecv (ack)");
					/* check the error and try to decide what to do */
					a_ptr_rem(j_p->al_p, a_p);
					errno = myerr;
					return(-1);
				}
				else if ((a_p->u.ack.size -= comp) > 0) /* partially completed */ {
					LOG("partial\n");
					a_p->u.ack.cur_p += comp;
					return(0);
				}
				LOG("done!\n");

				/* check if requested data size > actual size for J_READ */
				if (j_p->type == J_READ
				&& j_p->psize[a_p->u.ack.iod_nr] >
					((iack_p)a_p->u.ack.ack_p)->dsize)
				{
					LOG("do_job: getting less data than expected\n");
					if (prune_alist(j_p->al_p, A_READ, a_p->sock,
					j_p->psize[a_p->u.ack.iod_nr],
					((iack_p)a_p->u.ack.ack_p)->dsize) < 0) {
						ERR("do_job: prune failed\n");
					}
				}
				/* if ack value smaller, prune the access list */
				a_ptr_rem(j_p->al_p, a_p); /* remove ack access from list */
				break;
			case A_REQ:
				LOG1("req (s = %d) ", a_p->sock);
				comp = nbsend(a_p->sock, a_p->u.req.cur_p, a_p->u.req.size);
				if (comp < 0) {
					myerr = errno;
					PERROR("do_job: nbsend (req)");
					/* check the error and try to decide what to do */
					a_ptr_rem(j_p->al_p, a_p);
					errno = myerr;
					return(-1);
				}
				else if ((a_p->u.req.size -= comp) <= 0) {
					LOG("done!\n");
					free(a_p->u.req.req_p); /* free memory */
					a_ptr_rem(j_p->al_p, a_p);
				}
				else {
					LOG("partial\n");
					a_p->u.req.cur_p += comp;
					return(0);
				}
				break;
			default:
				ERR1("do_job: invalid access type (%d)\n", a_p->type);
				errno = EINVAL;
				return(-1);
		} /* end of switch */
		/* finished this access; get the next one */
		if (a_p = a_search(j_p->al_p, sock)) /* more work for this socket */ {
			/* update socket set */
			if (oldtype != a_p->type) switch (a_p->type) {
				case A_ACK:
				case A_READ:
					delsock(sock, ss_p);
					addsockrd(sock, ss_p);
					return(0); /* always stop when switching to a read type req */
					break;
				case A_WRITE:
				case A_REQ:
					delsock(sock, ss_p);
					addsockwr(sock, ss_p);
					break;
				default:
					break;
			}
			/* see if the next access is the next one for this socket */
			if (a_get_start(j_p->al_p) != a_p) return(0);
			oldtype = a_p->type;
		}
		else /* done with this socket */ {
			delsock(sock, ss_p);
			if (alist_empty(j_p->al_p)) /* done w/job */ {
				/* job is removed one function depth up in do_jobs */
				return(0);
			}
			/* remove socket from job's set */
         FD_CLR(sock, &j_p->socks);
			if (sock + 1 >= j_p->max_nr) /* need to reduce max_nr */ {
				int i;
				for (i=sock; i >= 0; i--) if (FD_ISSET(i, &j_p->socks)) break;
				j_p->max_nr = i+1;
			}
		}
	} /* end of while(a_p) */
	return(0);
}















