/* mgrcomm.c copyright (c) 1998 Clemson University, all rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as published
	by the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Contact:  Rob Ross  	rbross@parl.eng.clemson.edu

/*/

/*
 * $Log: mgrcomm.c,v $
 * Revision 1.5  1999/06/17 15:36:46  rbross
 * *** empty log message ***
 *
 * Revision 1.4  1998/07/21 04:16:09  rbross
 * fixed bug - was closing mgr sockets when ack status was bad
 *
 * Revision 1.3  1998/07/14 18:40:47  rbross
 * *** empty log message ***
 *
 * Revision 1.1  1998/07/05 00:48:45  rbross
 * Initial revision
 *
 *
/*/

static char mgrcomm_c[] = "$Header: /projects/pvfs/pvfs-working/lib/RCS/mgrcomm.c,v 1.5 1999/06/17 15:36:46 rbross Exp $";

#include <lib.h>

extern fdesc_p pfds[];
extern int errno;

/* MGRINFO structure - this is used to keep up with managers in use
 *
 * The idea here is to try to keep a single connection open to each
 * manager that we are communicating with.  To do this, we're going to
 * keep a little table and do our matching by comparing the port and 
 * address that we would use to connect to the manager.  We're going
 * to hold a maximum of MAXMGRS connections around.  There is no reason
 * why this number can't be bigger.
 *
 */
#define MAXMGRS 5
static int mgrcnt = 0;
static struct mgrinfoitem {
	unsigned short int port;
	unsigned int addr;
	int fd;
} mgrinfo[MAXMGRS];

/* PROTOTYPES */
static int addmgrinfo(struct sockaddr *saddr_p, int fd);
static int getmgrfd(struct sockaddr *saddr_p);
static int badmgrfd(int fd);

/* SEND_MREQ() - send a request to the manager and receive an ack
 *
 * This function handles sending a request to the manager, including
 * opening a connection to the manager if one is not already
 * established.  If an error occurs during the process, the socket will
 * be closed.  If the acknowledgement from the manager indicates
 * success, an open socket will be returned to the calling process so
 * that the process can receive trailing data (if any) or keep the
 * socket open.
 *
 * Note: CURRENTLY NOT USED IN PVFS_OPEN()
 *
 * PARAMETERS:
 * dir_p  - directory metadata, only needed if no fd passed in
 * req_p  - pointer to request structure
 * data_p - pointer to data to send following request (may be NULL)
 * ack_p  - pointer to memory location in which to store ack from mgr
 */
int send_mreq_dmeta(dmeta_p dir_p, mreq_p req_p, void *data_p,
                    mack_p ack_p)
{
	struct sockaddr saddr;

	if (!req_p || !ack_p) {
		errno = EINVAL;
		return(-1);
	}

	/* build a sockaddr structure */
	if (!dir_p || init_sock(&saddr, dir_p->host, dir_p->port)) return(-1);

	/* call send_mreq_saddr() to do the work */
	return(send_mreq_saddr(&saddr, req_p, data_p, ack_p));
} /* end of send_mreq_dmeta() */


/* SEND_MREQ_SADDR() - same as SEND_MREQ_DMETA, only pulls connection
 * info from a sockaddr structure instead.
 */
int send_mreq_saddr(struct sockaddr *saddr_p, mreq_p req_p,
                    void *data_p, mack_p ack_p)
{
	int fd;
	char newconn=0;

	if (!req_p || !ack_p || !saddr_p) {
		errno = EINVAL;
		return(-1);
	}
make_new_conn:
	if ((fd = getmgrfd(saddr_p)) < 0) {
		if ((fd = new_sock()) < 0) return(-1);
mgrcomm_connect_retry:
		if (connect(fd, saddr_p, sizeof(struct sockaddr))<0) {
			if (errno == EINTR) goto mgrcomm_connect_retry;
			syscall(SYS_close, fd);
			return(-1);
		}
		newconn=1;
	}
	/* send request */
	req_p->majik_nr = MGR_MAJIK_NR;
	if (bsend(fd, req_p, sizeof(mreq)) < sizeof(mreq)) {
		PERROR("send_mreq_saddr (sending req)");
		syscall(SYS_close, fd);
		if (!newconn) {
			badmgrfd(fd);
			goto make_new_conn; /* try to reopen connection */
		}
		return(-1);
	}
	/* send data (if any) */
	if (req_p->dsize > 0 && data_p
	&& bsend(fd, data_p, req_p->dsize) < req_p->dsize) {
		PERROR("send_mreq_saddr (sending data)");
		syscall(SYS_close, fd);
		if (!newconn) badmgrfd(fd);
		return(-1);
	}
	/* receive ack */
	if (brecv(fd, ack_p, sizeof(mack)) < sizeof(mack)) {
		syscall(SYS_close, fd);
		if (!newconn) badmgrfd(fd);
		return(-1);
	}
	if (ack_p->majik_nr != MGR_MAJIK_NR) /* serious error -- nonsense ack */ {
		syscall(SYS_close, fd);
		if (!newconn) badmgrfd(fd);
		errno = ack_p->eno;
		return(-1);
	}
	if (ack_p->status) /* status error -- don't close the connection...  */ {
		errno = ack_p->eno;
		return(-1);
	}
	if (newconn) addmgrinfo(saddr_p, fd);
	return(fd);
} /* end of send_mreq_saddr() */

static int addmgrinfo(struct sockaddr *saddr_p, int fd)
{
	if (mgrcnt >= MAXMGRS) {
		ERR("addmgrinfo: MAXMGRS exceeded!\n");
		return(-1);
	}
	mgrinfo[mgrcnt].addr = ((struct sockaddr_in *) saddr_p)->sin_addr.s_addr;
	mgrinfo[mgrcnt].port = ((struct sockaddr_in *) saddr_p)->sin_port;
	mgrinfo[mgrcnt].fd = fd;
	mgrcnt++;

	/* update fd table */
	if (pfds[fd]) {
		ERR1("fd %d already in use?!?\n", fd);
	}
	pfds[fd] = (fdesc_p)malloc(sizeof(fdesc));
	pfds[fd]->fs = FS_RESV;
	pfds[fd]->part_p = NULL;
	pfds[fd]->dim_p  = NULL;
	pfds[fd]->fd.ref = -1;
	pfds[fd]->fd.grp_nr = -1;
	return(0);
}

static int getmgrfd(struct sockaddr *saddr_p)
{
	int i=0;
	for (; i < mgrcnt; i++) {
		if (((struct sockaddr_in *)saddr_p)->sin_addr.s_addr == mgrinfo[i].addr
		&& ((struct sockaddr_in *)saddr_p)->sin_port == mgrinfo[i].port)
		{
			return(mgrinfo[i].fd);
		}
	}
	return(-1);
}

static int badmgrfd(int fd)
{
	int i=0, found=-1;

	for (; i < mgrcnt; i++) {
		if (fd == mgrinfo[i].fd) found=i;
	}
	if (found < 0) return(-1);
	if (found != mgrcnt-1) {
		mgrinfo[found] = mgrinfo[mgrcnt-1];
	}
	mgrcnt--;

	/* update fd table */
	if (!pfds[fd] || pfds[fd]->fs != FS_RESV) {
		ERR1("fd %d was not reserved?!?\n", fd);
		return(-1);
	}
	free(pfds[fd]);
	pfds[fd] = NULL;
	return(0);
}
