/*
 * mgr_config.c copyright (c) 1998 Clemson University, all rights reserved.
 *
 * Written by Rob Ross.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.eng.clemson.edu
 */

/*
 * These functions parse the config file for the mgr daemon and set
 * up the daemon's environment based on the specified configuration.
 * 
 * File format is simple:
 * 
 * 1) Blank lines are ignored
 * 2) Lines starting with '#' are ignored
 * 3) All other lines are of the format <option> <value>, where value
 *    is an arbitrary string ended by a carriage return or '#'
 * 4) In most cases only the first whitespace separated item in the
 *    value field is used for setting the option (eg. port).
 * 
 * SAMPLE CONFIG FILE:
 * 
 * #
 * # MGR configuration file
 * #
 * 
 * port 6969
 * user pvfs
 * rootdir /
 * debug 0
 * 
 * END OF SAMPLE CONFIG FILE
 *
 * Separate functions are used for reading the configuration file and
 * setting up the resulting environment.
 *
*/

#include <stdio.h>
#include <string.h>
#include <grp.h>
#include <pwd.h>
#include <sys/types.h>

#define INBUFSZ 1024
#define MAXOPTLEN 1024

static struct mgr_config {
	unsigned short port;
	char user[MAXOPTLEN];
	char group[MAXOPTLEN];
	char rootdir[MAXOPTLEN];
	int debug;
} config = { 7000, "root", "root", "/", 0 };


#if 0
int main()
{
	char wd[1024];
	parse_config("iod.conf", "iod");
	dump_config(stdout);
	set_config();

	printf("uid = %d, gid = %d, cwd = %s\n", getuid(), getgid(),
	getcwd(wd, 1024));
}
#endif

int parse_config(char *fname)
{
	FILE *cfile;
	char inbuf[INBUFSZ], *name, *option, *value;

	/* open file */
	if (!(cfile = fopen(fname, "r"))) {
		perror("fopen");
	}

	while (fgets(inbuf, INBUFSZ, cfile)) {
		/* comments get skipped here */
		if (*inbuf == '#') continue;

		/* blank lines get skipped here */
		if (!(option = strtok(inbuf, " \t\n"))) continue;

		if (!(value = strtok(NULL, "#\n"))) {
			fprintf(stderr, "parse_config: error tokenizing value\n");
			return(-1);
		}
		/* PORT (eg. "port 7000") */
		if (!strcmp("port", option)) {
			char *err;
			if (!(strtok(value, " \t\n#"))) {
				fprintf(stderr, "parse_config: error reducing string");
				continue;
			}
			config.port = strtol(value, &err, 10);
			if (*err) /* bad character in value */ {
				fprintf(stderr, "warning: trailing character(s) in port\n");
			}
		}
		/* USER (eg. "user nobody") */
		else if (!strcmp("user", option)) {
			if (!(strtok(value, " \t\n#"))) {
				fprintf(stderr, "parse_config: error reducing string");
				continue;
			}
			strncpy(config.user, value, MAXOPTLEN);
		}
		/* GROUP (eg. "group bin") */
		else if (!strcmp("group", option)) {
			if (!(strtok(value, " \t\n#"))) {
				fprintf(stderr, "parse_config: error reducing string");
				continue;
			}
			strncpy(config.group, value, MAXOPTLEN);
		}
		/* ROOTDIR (eg. "rootdir /tmp") */
		else if (!strcmp("rootdir", option)) {
			if (!(strtok(value, " \t\n#"))) {
				fprintf(stderr, "parse_config: error reducing string");
			}
			strncpy(config.rootdir, value, MAXOPTLEN);
		}
		/* DEBUG (eg. "debug 4") */
		else if (!strcmp("debug", option)) {
			char *err;
			if (!(strtok(value, " \t\n#"))) {
				fprintf(stderr, "parse_config: error reducing string");
			}
			config.debug = strtol(value, &err, 10);
			if (*err) /* bad character in value */ {
				fprintf(stderr, "warning: trailing character(s) in debug\n");
			}
		}
		else {
			fprintf(stderr, "unknown option: %s\n", option);
		}
	}
	fclose(cfile);
	return(0);
} /* end of parse_config() */

int dump_config(FILE *fp)
{
	fprintf(fp, "# MGR DAEMON CONFIG FILE -- AUTOMATICALLY GENERATED\n");
	fprintf(fp, "port %d\n", config.port);
	fprintf(fp, "user %s\n", config.user);
	fprintf(fp, "group %s\n", config.group);
	fprintf(fp, "rootdir %s\n", config.rootdir);
	fprintf(fp, "debug %d\n", config.debug);
	return(0);
} /* end of dump_config() */

/* SET_CONFIG() - set environment using configuration parameters
 *
 * Steps:
 * 1) find gid, uid
 * 2) Set current working directory to rootdir
 * 3) chroot() to rootdir
 * 4) set gid, uid
 *
 */
int set_config(void)
{
	int uid, gid; 
	struct group *grp_p;
	struct passwd *pwd_p;

	if (!(grp_p = getgrnam(config.group))) {
		perror("set_config: getgrnam");
		return(-1);
	}
	gid = grp_p->gr_gid;

	if (!(pwd_p = getpwnam(config.user))) {
		perror("set_config: getpwnam");
		return(-1);
	}
	uid = pwd_p->pw_uid;

	if (chdir(config.rootdir) < 0) {
		perror("set_config: chdir");
		return(-1);
	}

	if (chroot(config.rootdir) < 0) {
		perror("set_config: chroot");
		return(-1);
	}

	if (setgid(gid) < 0) {
		perror("set_config: setgid");
		return(-1);
	}

	if (setuid(uid) < 0) {
		perror("set_config: setuid");
		return(-1);
	}
	return(0);
} /* end of set_config() */

int get_config_port(void)
{
	return(config.port);
}

int get_config_debug(void)
{
	return(config.debug);
}
