/*
 * check_pvfs.c copyright (c) 1997 Clemson University, all rights reserved.
 *
 * Written by Matt Cettei and Walt Ligon.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross	 rbross@parl.eng.clemson.edu
 *			  Matt Cettei mcettei@parl.eng.clemson.edu
 */

/*
 * CHECK_PVFS.C - determine if a file or directory is a PVFS file/dir.
 *
 */

/*
 * $Log: check_pvfs.c,v $
 * Revision 1.8  1998/07/09 21:11:16  rbross
 * modified to identify .pvfsdir and .iodtab files and return 0
 *
 * Revision 1.6  1998/07/02 18:33:15  rbross
 * modified to return 2 for PVFS directories
 *
 * Revision 1.5  1998/02/21 03:29:16  rbross
 * yet another stat fix
 *
 * Revision 1.4  1998/02/19 23:03:06  rbross
 * removed SYS_stat defines ... moving to lib.h
 *
 * Revision 1.3  1998/02/16 22:03:56  rbross
 * repatched to work with libc5 too
 *
 * Revision 1.2  1998/02/16 18:55:18  rbross
 * Newt patches
 *
 * Revision 1.1  1997/01/16 16:30:50  mcettei
 * Initial revision
 *
 *
/*/

static char check_pvfs_c[] = "$Header: /projects/pvfs/pvfs-working/shared/RCS/check_pvfs.c,v 1.8 1998/07/09 21:11:16 rbross Exp $";

#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/timeb.h>
#include <syscall.h>
#include <fcntl.h>
#include <unistd.h>
#include <meta.h>
#include <req.h>

extern int errno;

/* check_pvfs() - checks to see if the filename pointed to by "pathname"
 * is or would be a PVFS file.  This is determined as follows.  If the
 * name refers to a file that exists or a name that does not exist, the
 * directory is checked for the existance of a ".pvfsdir" file.  If this
 * file is there, the file is (or would be) a PVFS file and 1 is
 * returned.
 *
 * If there is no ".pvfsdir" file, 0 is returned.
 * If there is a ".pvfsdir" file and the pathname refers to a directory,
 * 2 is returned.
 */

#define MAXTMP 512
int check_pvfs(const char* pathname)
{
	int length, isdir=0;
	char tmp[MAXTMP];
	static char pvfsdir[]=".pvfsdir", iodtab[]=".iodtab";
	struct stat foo;

	if (!pathname) {
		errno = EFAULT;
		return(-1);
	}

	strncpy(tmp, pathname, MAXTMP);
	length = strlen(tmp);
	if (length == 0) {
		errno = ENOENT;
		return(-1);
#if 0
		/* this should never happen - WBL */
		strncat(tmp, ".", MAXTMP);
#endif
	}
	if (unix_stat(tmp, &foo) < 0) {
		/* file doesn't exist.  We need to fool the next check so that we
		 * can see if this would be a PVFS file were it to be created...
		 */
		foo.st_mode = 0;
	}
	if (!S_ISDIR(foo.st_mode)) {
#if 0
		/* strip off the file name */
		for (; length >= 0 && tmp[length] != '/'; tmp[length--] = 0);
#endif
		/* find division between file name and directory */
		for (; length >= 0 && tmp[length] != '/'; length--);

		/* check for names that are used by PVFS */
		/* return 0 (UNIX file) for these */
		if (!strcmp(pvfsdir, &tmp[length+1])
		|| !strcmp(iodtab, &tmp[length+1])) return(0);

		if (length < 0) /* no directory here! */ {
			tmp[0] = '.';
			tmp[1] = 0;
		}
		else  /* cut off the file name */ {
			tmp[length] = 0;
		}
	}
	else {
		isdir=1;
#if 0
		/* add on a "/" in case it isn't already there */
		strncat(tmp, "/", MAXTMP);
#endif
	}
#if 0
	strncat(tmp, ".pvfsdir", MAXTMP);
#endif
	strncat(tmp, "/.pvfsdir", MAXTMP);
	if (unix_stat(tmp, &foo) < 0) {
		return(0);
	}
	else {
		/* Otherwise must be pvfs file */
		return(1+isdir);
	}
}






