/*
/* sockio.c copyright (c) 1994 Clemson University, all rights reserved.
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as published
	by the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	Contact:  Rob Ross  	rbross@parl.eng.clemson.edu
				 Matt Cettei mcettei@parl.eng.clemson.edu

/*/

/*
/* SOCKIO.C - socket routines
/*
/* Thanks go to Aric Blumer & Walt Ligon; they did most of the work.
/*
/*/

/*
 * $Log: sockio.c,v $
 * Revision 1.9  1999/02/19 00:29:50  rbross
 * now we're EINTR friendly
 *
 * Revision 1.8  1999/02/18 19:33:19  rbross
 * defined out dump_sockaddr...conflicts with function in mpich
 *
 * Revision 1.7  1998/02/16 19:52:04  rbross
 * Newt patches
 *
 * Revision 1.6  1997/10/28 19:37:03  pvfs
 * added license
 *
 * Revision 1.5  1997/06/23 20:22:46  rbross
 * added set_tcpopt() call (mainly to set/clear Nagle algorithm)
 *
 * Revision 1.4  1997/06/23 20:16:45  rbross
 * *** empty log message ***
 *
 * Revision 1.3  1997/02/21 16:45:03  rbross
 * brecv() correctly returns 0 when socket is closed now
 *
 * Revision 1.2  1997/02/06 16:08:52  rbross
 * fixed BIG bug in all the send/recv's
 *
 * Revision 1.1  1996/11/25 22:37:14  rbross
 * Initial revision
 *
/*/

static char sockio_c[] = "$Version$";

/* UNIX INCLUDE FILES */
#include <errno.h>
#include <stdio.h>
#include <string.h>	/* bzero and bcopy prototypes */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <debug.h>

/* PVFS INCLUDE FILES */
#include <sockio.h>

extern int errno;

/* FUNCTIONS */
int new_sock() {
	static int p_num = -1; /* set to tcp protocol # on first call */
	struct protoent *pep;

   if (p_num == -1) {
		if ((pep = getprotobyname("tcp")) == NULL) {
			perror("Kernel does not support tcp");
			return(-1);
		}
		p_num = pep->p_proto;
	}
	return(socket(AF_INET, SOCK_STREAM, p_num));
}

int bind_sock(int sockd, int service)
{
	struct sockaddr_in saddr;

	bzero((char *)&saddr,sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons((u_short)service);
	saddr.sin_addr.s_addr = INADDR_ANY;
bind_sock_restart:
	if (bind(sockd,(struct sockaddr *)&saddr,sizeof(saddr)) < 0) {
		if (errno == EINTR) goto bind_sock_restart;
		return(-1);
	}
	return(sockd);
}

int connect_sock(int sockd, char *name, int service)
{
	struct sockaddr saddr;

	if (init_sock(&saddr, name, service) != 0) return(-1);
connect_sock_restart:
	if (connect(sockd,(struct sockaddr *)&saddr,sizeof(saddr)) < 0) {
		if (errno == EINTR) goto connect_sock_restart;
		return(-1);
	}
	return(sockd);
}

int init_sock(struct sockaddr *saddrp, char *name, int service)
{
	struct hostent *hep;

	bzero((char *)saddrp,sizeof(struct sockaddr_in));
	if (name == NULL) {
		if ((hep = gethostbyname("localhost")) == NULL) {
			return(-1);
		}
	}
	else if ((hep = gethostbyname(name)) == NULL) {
		return(-1);
	}
	((struct sockaddr_in *) saddrp)->sin_family = AF_INET;
	((struct sockaddr_in *)saddrp)->sin_port = htons((u_short)service);
	bcopy(hep->h_addr, (char *)&(((struct sockaddr_in *)saddrp)->sin_addr),
		hep->h_length);
	return(0);
}

/* blocking receive */
/* Returns -1 if it cannot get all len bytes
 * and the # of bytes received otherwise
 */
int brecv(int s, void *buf, int len)
{
	int oldfl, ret, comp = len;
	oldfl = fcntl(s, F_GETFL, 0);
	if (oldfl & O_NONBLOCK) fcntl(s, F_SETFL, oldfl & (~O_NONBLOCK));

	while (comp) {
brecv_restart:
		if ((ret = recv(s, (char *) buf, comp, 0)) < 0) {
			if (errno == EINTR) goto brecv_restart;
			return(-1);
		}
		if (!ret) {
			/* Note: this shouldn't be happening since we should block,
			 * but in the case where the other end of the socket has been
			 * closed we are seeing this behavior...
			 *
			 * New note: this indicates a closed socket.  However, I don't
			 * like this behavior, so we're going to return -1 w/an EPIPE
			 * instead.
			 */
			errno = EPIPE;
			return(-1);
		}
		comp -= ret;
		buf += ret;
	}
	return(len - comp);
}

/* nonblocking receive */
int nbrecv(int s, void *buf, int len)
{
	int oldfl, ret, comp = len;

	oldfl = fcntl(s, F_GETFL, 0);
	if (!(oldfl & O_NONBLOCK)) fcntl(s, F_SETFL, oldfl | O_NONBLOCK);

	while (comp) {
nbrecv_restart:
		ret = recv(s, buf, comp, 0);
		if (!ret) /* socket closed */ {
			errno = EPIPE;
			return(-1);
		}
		if (ret == -1 && errno == EWOULDBLOCK) {
			LOG("nbrecv: would block\n");
			return(len - comp); /* return amount completed */
		}
		if (ret == -1 && errno == EINTR) {
			goto nbrecv_restart;
		}
		else if (ret == -1) {
			PERROR("nbrecv: recv");
			return(-1);
		}
		LOG3("nbrecv: s = %d, expect = %d, actual = %d\n", s, comp, ret);
		comp -= ret;
		buf += ret;
	}
	return(len - comp);
}

/* blocking send */
int bsend(int s, void *buf, int len)
{
	int oldfl, ret, comp = len;
	oldfl = fcntl(s, F_GETFL, 0);
	if (oldfl & O_NONBLOCK) fcntl(s, F_SETFL, oldfl & (~O_NONBLOCK));

	while (comp) {
bsend_restart:
		if ((ret = send(s, (char *) buf, comp, 0)) < 0) {
			if (errno == EINTR) goto bsend_restart;
			return(-1);
		}
		comp -= ret;
		buf += ret;
	}
	return(len - comp);
}

/* nonblocking send */
/* should always return 0 when nothing gets done! */
int nbsend(int s, void *buf, int len)
{
	int oldfl, ret, comp = len;
	oldfl = fcntl(s, F_GETFL, 0);
	if (!(oldfl & O_NONBLOCK)) fcntl(s, F_SETFL, oldfl | O_NONBLOCK);

	while (comp) {
nbsend_restart:
		ret = send(s, (char *) buf, comp, 0);
		if (ret == 0 || (ret == -1 && errno == EWOULDBLOCK))
			return(len - comp); /* return amount completed */
		if (ret == -1 && errno == EINTR) {
			goto nbsend_restart;
		}
		else if (ret == -1) return(-1);
		comp -= ret;
		buf += ret;
	}
	return(len - comp);
}

/* routines to get and set socket options */
int get_sockopt(s, optname)
int s, optname;
{
	int val, len = sizeof(val);
	if (getsockopt(s, SOL_SOCKET, optname, &val, &len) == -1)
		return(-1);
	else return(val);
}

int set_tcpopt(s, optname, val)
int s, val, optname;
{
	if (setsockopt(s, SOL_TCP, optname, &val, sizeof(val)) == -1)
		return(-1);
	else return(val);
}

int set_sockopt(s, optname, size)
int s, size, optname;
{
	if (setsockopt(s, SOL_SOCKET, optname, &size, sizeof(size)) == -1)
		return(-1);
	else return(size);
}

int get_socktime(s, optname)
int s, optname;
{
	struct timeval val;
	int len = sizeof(val);
	if (getsockopt(s, SOL_SOCKET, optname, &val, &len) == -1)
		return(-1);
	else return(val.tv_sec * 1000000 + val.tv_usec);
}

int set_socktime(s, optname, size)
int s, size, optname;
{
	struct timeval val;
	val.tv_sec = size / 1000000;
	val.tv_usec = size % 1000000;
	if (setsockopt(s, SOL_SOCKET, optname, &val, sizeof(val)) == -1)
		return(-1);
	else return(size);
}

#if 0
/* DUMP_SOCKADDR() - dump info in a sockaddr structure
/*
/* Might or might not work for any given platform!
/*/
int dump_sockaddr(ptr, fp)
struct sockaddr_in *ptr;
FILE *fp;
{
	int i;
	unsigned long int tmp;
	struct hostent *hp;
	char abuf[]="xxx.xxx.xxx.xxx\0";

	fprintf(fp, "sin_family = %d\n", ptr->sin_family);
	fprintf(fp, "sin_port = %d (%d)\n", ptr->sin_port, ntohs(ptr->sin_port));
	/* print in_addr info */
	tmp = ptr->sin_addr.s_addr;
	sprintf(&abuf[0], "%d.%d.%d.%d",tmp & 0xff, (tmp >> 8) & 0xff,
		(tmp >> 16) & 0xff, (tmp >> 24) & 0xff);
	hp = gethostbyaddr((char *)&ptr->sin_addr.s_addr, sizeof(ptr->sin_addr),
		ptr->sin_family);
	fprintf(fp, "sin_addr = %lx (%s = %s)\n", ptr->sin_addr, abuf, 
		hp->h_name);
	for (i = 0; i < sizeof(struct sockaddr) - sizeof(short int) -
		sizeof(unsigned short int) - sizeof(struct in_addr); i++)
	{
		fprintf(fp, "%x", ptr->sin_zero[i]);
	}
	fprintf(fp, "\n");
} /* end of DUMP_SOCKADDR() */
#endif
